#!/usr/bin/perl -w

=head1 NAME

mkbkup.pl - to make a bootable backup of your Linux OS + all mounted file systems.

=head1 DESCRIPTION

Run "%perl mkbkup.pl -help" to find out command-line options.


$Date: 2002/11/08 17:20:10 $ by Newton Hammet (newton@hammet.net )

=head1 PREREQUISITES

Getopt::Long;;
Term::ANSIColor qw(:constants);

=head1 COREQUISITES

=pod SUPPORTED OS

Linux (all versions that support the 'sfdisk' command)

=pod SCRIPT CATEGORIES

CPAN/Administrative
CPAN

=head1 USAGE

=head2 Example 1

	mkbkup.pl /dev/hdg

In this example all partitions on the backup drive will be 
identical to partitions on the current OS.  The backup OS
drive is /dev/hdg.

=head2 Example 2

	mkbkup.pl -nupart /=10G,/home=20G /dev/hdg

In this example except for boot and swap, all other partitions
on current OS will be ignored and new partitions will be created
for '/' at 10 Gigabytes, and '/home' at 20 gigabytes.  The backup
copy is placed on /dev/hdg.

=head1 SOFTWARE LICENSE

Copyright (c) 2002, Newton Hammet
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.
* Neither the name of the  mkbkup.pl-system nor the names of any of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

=cut

#
# Global Constants
#
my $VERSION = "2.0";
my $MAXMOUNTS = 60;					# Maximum number of mount points allowed.
my $MAXLEVELS = 10;					# Maximum subdirectory levels for mounted fs.
my $DEF_FSTYP = "ext3";				# Default Filesystem type for new partitions.
#
# Disk sector size : Change only '$bytespec'
#                    ($megabpsec & $gigabsec are calculated.)
#
my $bytespsec = 512;                # bytes per sector
my $megabpsec = 1048576/$bytespsec; # megabytes per sector
my $gigabpsec = $megabpsec * 1024;	# gigabytes per sector
#
# Global Arrays
#
my @partinp = (); # 1 element per disk partition stanza (nupart option)
my @partbld = (); # 1 element per disk partition, sfdisk input.
my @hdd   = ();   # 1 element per hard drive device (the device name)
my @dffs  = ();   # partition name for each mounted fs on current os
my @sorted= ();   # same as above but sorted in partition number order
my @mpts  = ();   # for each '@dffs' entry the corresponding mount point
my @numpts  = (); # boot mount point for each mounted fs on backup os
my @cmdout = ();  # stdout for any system command fm 'runme' goes here.
#
# Global hash tables
#
my %hddsz = ();     # size in kilobytes of each hard drive,     key=device-name(/dev/hdX)
my %hddcyl = ();    # number of cylinders for each hard drive,  key=device-name
my %bytpcyl = ();   # bytes per cylinder for each hard drive,   key=device-name
my %secpcyl = ();   # sectors per cylinder for each hard drive, key=device-name
my %hddhds = ();    # tracks per cylinder for each hard drive,  key=device-name
my %hddsec = ();    # sectirs per track for each hard drive,    key=device-name
my %mp2fs = ();     # each mounted fs,         key=mount-point     value=partition-name
my %fs2fstyp = ();  # each mounted fs,         key=partition-name  value=fs-type
my %nump2fs = ();   # each fs on backup drive, key=mount-point     value=partition-name
my %nufs2fstyp = ();# each fs on backup drive, key=partition-name  value=fs-type
my %nufstyp = ();   # each fs on backup drive, key=mount-point     value=fs-type
my %nufscap = ();   # each fs on backup drive, key=mount-point     value=fs-capacity-kilobytes
my %nulbl = ();     # Created volume label for each ext3 fs on backup drive, key=mount-point
my %fs2mp = ();     # each mountpoint on current os, key=partition-name, value=mount-point
my %fscap = ();     # each mountpoint on current os, key=mount-point,  value=capacity
my %nolbls= ();     # key=fstype,  value="NOLABEL"
my %contains_fs= ();# key=full-path, value=full-path-to-actual-mountpoint
#
# system command stanzas for sfdisk, df 
#
my $GetHDsize  =  "/sbin/sfdisk -s "; # Get HDD size in # of KB
my $GetHDgeom  =  "/sbin/sfdisk -g "; # Get HDD geometry in cyl, heads, sect/trk
my $GetHDpart  =  "/sbin/sfdisk -d "; # Get HDD partition table.
my $PutHDpart  =  "/sbin/sfdisk ";    # Partition a HDD (be careful!)
my $GetDf   =  "/bin/df";             # The df command.
#
# === Command line parameters and options
#
my $opt_verbose;
my $opt_debug;
my $opt_newpart;
my $opt_order = 0;
my $opt_output;
my $opt_nolabels;
my $bkup_os;
#
# === Other global variables determined at runtime
#
my $curr_os;
my $bkup_mtpt;
my $devln;

sub usage {
	print <<EOF_USAGE;
Usage:
perl mkbkup.pl OPTIONS BackupHDaddress
OPTIONS:
    [-debug] default is false
    [-verbose] default is false
    [--newpart part1,part2,...] where: part1, part2, are
                                strings of the following format:

<mountpoint>=<size>/<parttyp>/[boot]

<mountpoint> =s  beginning with '/', or the word 'swap'.
                word "swap" to indicate a swap partition.
<size>        in gigabytes(g), megabytes(m), or sectors(s),
              (1 sector = 512 bytes).

Sample OPTIONS (--newpart option) syntax:
	--newpart swap=1g,/=10g/83,/home=12g/83

	(which assigns as new partitions, 1 gigabyte to swap,
	10 gigabytes to '/', and 20 gigabytes to '/home'.)

BackupHDaddress:
	Any device name as: '/dev/hdx', 'dev/hdx', 'hdx' , where

EXAMPLE:
    cmdln> mkbkup.pl /dev/hdg
NOTE:
	creates bootable backup, with partitions matching the
    current partitions in order, size, and fs type.
	(called the "verbatim" option)
EOF_USAGE
exit;
} ############# END OF USAGE #########################

# Pexit: provide an error message and exit the program,
#   because 'die' is not quite graceful enough.
#
sub Pexit {
my $errmsg = $_[0];
print "\nERROR in $0, VERSION: $VERSION\n";
print $errmsg;
print "\n\n";
exit 1;
}

# getdir =s reads everybody, incl. dot files, but not the '.' and the '..'
#
# Input:  1 Parameter or null; if null then pathname is '/'.
#         the 1 parameter is the pathname of a directory.
# Output: an array containing all the filenames, included dot filenames
#         but not the '.' and '..' filenames.
#
sub getdir {
my $path = $_[0];
opendir (DIRECTORY, $path) or Pexit "ERR01: Could not open current dir";
my @dirs = (readdir(DIRECTORY));
shift @dirs;
shift @dirs;
closedir (DIRECTORY);
return (sort @dirs);
}

# Print the options that came from the command line.
#
#
sub print_params {
if (defined $opt_verbose) {
	print "\nOptions in effect:\n";
	if(defined $opt_debug){print "debug is defined. debug is verbose and then some\n"}
	if(defined $opt_newpart){print "new partitions = $opt_newpart\n"}
	else {print "No new partitions, use existing partitions.\n"}
	if (defined $bkup_os) {print "backup disk = $bkup_os\n"}
	else {print "No backup disk specified.\n";}
	print "\n";
}
}

sub chmode {
sub rwx {
my $val = 0;
if ($_[0] =~ m/r/) {$val+=4};
if ($_[0] =~ m/w/) {$val+=2};
if ($_[0] =~ m/[stx]/) {$val+=1};
return $val;
}
my $val = 0;
my $rawperm = $_[0];
if (length($rawperm) ne 10) {Pexit "ERR02: invalid parameter to chmode"}
$x1 = rwx substr $rawperm, 1, 3;
$x2 = rwx substr $rawperm, 4, 3;
$x3 = rwx substr $rawperm, 7, 3;
if (substr($rawperm,3,1) =~ m/[sS]/) {$val+=4}
if (substr($rawperm,6,1) =~ m/[sS]/) {$val+=2}
if (substr($rawperm,9,1) =~ m/[tT]/) {$val+=1}
return "$val$x1$x2$x3";
}

sub prt_hdd_tbl {
my $devln;
print "===> Hard-drive geometries\n";
foreach $devln (@hdd) {
	my $act_size = $hddcyl{$devln} * $bytpcyl{$devln};
	print "$devln  $hddcyl{$devln} cylinders  $bytpcyl{$devln} bytes/cyl  $act_size total bytes\n";
}
}
#
# For mount points with volume labels print the new volume labels with each mount point
#
sub prt_nulbl {
my $mpt;
print "\n==== New labels for partitions on $bkup_os (backup drive)\n";
foreach $mpt (@numpts) {
	next unless (exists($nulbl{$mpt}));
	print "label on $bkup_os for $mpt\tis\t$nulbl{$mpt}\n";
}
}

sub prt_partbld {
my $thisline;
foreach $mpt (@partbld) {
	$thisline = "$mpt\n";
	$thisline =~ s/\n{1,99}/\n/g;
	print "$thisline";
}
}

sub prt_numpts {
my $mpt;
print "========= all of the pretty @numpts mount-points ===========\n";
foreach $mpt (@numpts) {
	print "$mpt\t$nump2fs{$mpt}\n";
}
}

sub prt_partinp {
my $mpt;
print "========= all of the pretty partinp stanzas ===========\n";
foreach $mpt (@partinp) {
	print "$mpt\n";
}
}

sub prt_cmdout {
my $mpt;
foreach $mpt (@cmdout) {
	print "$mpt";
}
}

sub prt_dffs {
my $dfent;
print "\n==== file systems that are currently mounted ===\n";
foreach $dfent (@dffs) {
	print "$fs2mp{$dfent}\t$dfent\n";
}
}
sub dirname {
my @dirout = `dirname $_[0] 2>&1`;
Pexit "ERR03: Could not find parent of $_[0]" unless (!$?);
my $parentdir = $dirout[0];
chomp $dirout[0];
return $dirout[0];
}

# runme:  run a system command and capture output.
# Input: positional parameter 1 is the command + args to be run.
#     the input can be 1 or more shell commands separated, by ';'.
# Output: return as a value the 1st line of the output from running
#     the command.
# If the command terminates with a nonzero error code then terminate
# the program after displaying all combined output from the command.
#
sub runme {
my $syscmd = $_[0];
@cmdout = `$syscmd 2>&1`;
if ($?) {
	my $rc = $?;
	while ($pparm = shift @_) {
		print "$pparm\n";
	}
	print "====combined stderr and stdout====\n";
	foreach $outln (@cmdout) {
		(length($outln)>2) && print "$outln";
	}
	Pexit "ERR04: '$syscmd' terminated abnormally (rc=$rc)";
} else {
	return $cmdout[0];
}
}

# roundup: return a whole number as a value.
# Input: 1st positional parameter is a number that
#    can have a decimal portion.
# Output: if the input is a whole number return that value.
#    if the input has a decimal portion drop the decimal portion
#    and return the integer portion + 1.
#
sub roundup {
my $inp = $_[0];
return $inp unless ($inp =~ m/(\d+)\.(\d+)/);
$inp =~ s/(\d+)\.(\d)(.*)/$1/;
return ($inp+1);
}

# get_boot_part ... fetch boot partition info (deprecated)
#
#  Return, as a value, a string with slash delimited values
#  of size, offset, and the word "boot".  This used to be
#  input to the "parse_nupart" routine but is used no longer.
#
sub get_boot_part {
	my ($junk, $off, $sz, $typ);
	runme "$GetHDpart $curr_os|grep \"^$curr_os\"",
		  "ERR05: $GetHDpart failure in 'sub get_boot_part'";
	$cmdout[0] =~ s/[:=, \t]{1,99}/ /g;
	($junk, $junk, $off, $junk, $sz, $junk, $typ) = split / /, $cmdout[0];
	return "$sz"."s/"."$typ/$off"."s/boot\n";
}

# parse_nupart ... Parse a list of disk partition descriptions
#
#  Parse the comma-separated string of new partition descriptions which
#  occurs following the command-line option, '--newpart'.  The disk
#  partition part, which occurs to the right of the '=' sign in each
#  disk partition stanza, is stored in the global array '@partinp'.
#  The mount point part which occurs to the right of the '=' sign in
#  each disk partition stanza, is stored in the global array '@numpts'.
#
# Input: $opt_newpart  Output: @numpts, @partinp.
#
sub parse_nupart {
my $partln;
my $NewPart;
my ($sz, $typ, $boot);
my $psz = "1G";
my @nupartlns = split /,/, $opt_newpart;
foreach $partln (@nupartlns) {
	$NewMpt = $partln;
	$NewMpt =~ s/(.+)=(.*)/$1/;
	if ($NewMpt eq $partln) {$partln .= "=/";}
	$Mpt = $NewMpt;
	$Mpt =~ s/(.+):(.*)/$1/;
	if ($Mpt eq $NewMpt) {$NewMpt .= ":DEFAULT";}
	push @numpts, $Mpt;
	$Typ = $NewMpt;
	$Typ =~ s/(.+):(.*)/$2/;
	$nufstyp{$Mpt} = $Typ;
	$NewMpt = $partln;
	$NewMpt =~ s/(.+)=(.*)/$2/;
	($sz, $typ, $boot) = split /\//, $NewMpt;
	if (!$boot) {$boot="";}
	if ($sz =~ m/^0+/) {$sz = "0G";}
	if (!$sz) {$sz = $psz;}
	else      {$psz = $sz;}
	if (!$typ) {
		if ($Mpt ne "swap") {$typ="83";}
		else {$typ="82";}
	}
	push @partinp, "$sz/$typ/$boot";
}
}

# build_disk_geometry_table === Build disk geometry table
#
# Input: All Harddrives seen by 'sfdisk'.
# Output: $hddcyl (total cylinders on each hard drive)
#       : $hddhds (total tracks/cyl, each drive)
#       : $hddsec (total sectors/track, each drive)
#		: $bytpcyl (total bytes/cylinder, each drive)
#		: $secpcyl (total sectors/cylinder, each drive)
#
sub build_disk_geometry_table {
my $dev;
my $devln;
my $sz;
my ($junk, $cyl, $heads, $sect);
runme "$GetHDgeom|grep '/dev'", "ERR06: something wrong fetching disk geometry";
foreach $devln (@cmdout) {
	chomp $devln;
	$dev = $devln;
	$dev =~ s/([\/a-z]+):.+/$1/g;
	$sz  = $devln;
	$sz  =~ s/([\/a-z]+):\s+(.+)\scyl(.+)/$2/g;
	$devln =~ s/\s+/,/g;
	$devln =~ s/,,/,/g;
	($junk, $cyl, $junk, $heads, $junk, $sect, $junk) = split /,/, $devln;
	$hddcyl{$dev} = $cyl;
	$hddhds{$dev} = $heads;
	$hddsec{$dev} = $sect;
	$bytpcyl{$dev} = $sect * $heads * $bytespsec;
	$secpcyl{$dev} = $sect * $heads;
}
if (!$hddcyl{$bkup_os}) {
	Pexit "ERR07:\tcould not find geometry for $bkup_os\n\tPossible invalid drive specification";
}
}

sub build_disk_size_table {
my $dev;
my $devln;
my $sz;
runme "$GetHDsize|grep '/dev'", "ERR08: something wrong fetching disk size";
foreach $devln (@cmdout) {
	chomp $devln;
	$dev = $devln;
	$dev =~ s/([\/a-z]+):.+/$1/g;
	$sz  = $devln;
	$sz  =~ s/([\/a-z]+):(.+)/$2/g;
	$hddsz{$dev} = $sz;
	push @hdd, $dev;
}
}

sub get_fstyp {
my $mpt;
my ($junk, $fstype);
runme "mount | grep $_[0]", "ERR09: called from get_fstyp";
$cmdout[0] =~ s/[ \t]{1,99}/ /g;
($junk, $junk, $junk, $junk, $fstype, $junk) = split / /, $cmdout[0];
return $fstype;
}

sub build_contains_fs_trail {
my $fpth = $_[0];
$fpth =~ s:/+(.+):/$1:;
$fpth =~ s:^([^/].+)$:/$1:;
my $mpt = $fpth;
while ($mpt ne "/") {
	$mpt = dirname $mpt;
	if ($mpt ne "/") {
		$contains_fs{"$mpt"} = $fpth;
	}
}
}

sub build_dev_mount_table_curr_os {
my $devln;
my ($dev, $cap, $used, $avl, $prct, $mpt);
runme "$GetDf|grep '^/dev'|sed 's/ \\{1,\\}/ /g'",
	  "ERR10: from build_dev_mount_table_curr_os";
my @blddf = @cmdout;
foreach $devln (@blddf) {
	chomp $devln;
	($dev, $cap, $used, $avl, $prct, $mpt) = split / /, $devln;
	push @dffs, $dev;
	push @mpts, $mpt;
	$mp2fs{$mpt} = $dev;
	$fs2fstyp{$dev} = get_fstyp $dev;
	$fs2mp{$dev} = $mpt;
	$fscap{$mpt} = $cap;
	$dev =~ s/(.+)(\d\d)/$1/;
	$dev =~ s/(.+)(\d)/$1/;
	if (!$curr_os) {
		$curr_os = $dev;
	} else {
		if ($dev ne $curr_os ) {
			Pexit "ERR11: OS Partition '$mp2fs{$mpt} = $mpt' on different hdd";
		}
	}
	build_contains_fs_trail $mpt;
	if ($#dffs > ($MAXMOUNTS-1)) {
		Pexit "ERR12: number of mount points > $MAXMOUNTS(maximum allowed)";
	}
}
}

sub inp2sec {
my $inp = $_[0];
my $spc = $_[1];
$inp =~ tr/[A-Z/[a-z]/;
if ($inp =~ m/(\d+)g/) {
	$inp =~ s/(\d+)g/$1/;
	return ($inp * $gigabpsec); }
if ($inp =~ m/(\d+)c/) {
	$inp =~ s/(\d+)c/$1/;
	return ($inp * $spc); }
if ($inp =~ m/(\d+)m/) {
	$inp =~ s/(\d+)m/$1/;
	return ($inp * $megabpsec); }
if ($inp =~ m/(\d+)s/) {
	$inp =~ s/(\d+)s/$1/;
	return ($inp); }
die "inp2sec: bad conversion value from arg1=$_[0], arg2=$_[1]";
}

# find_swap_curr_os :: find all partitions of type=0x82
#
# Input : @partbld
# Output: @dffs, %fs2mp
#
sub find_swap_curr_os {
	foreach $parln (@partbld) {
		if ($parln =~ m/Id=82/i) {
			$swap = $parln;
			$swap =~ s/(.+)\s(.+)/$1/;
			$swap =~ s/([\/a-z1-9]+)\s*:.+/$1/;
			chomp $swap;
			push @dffs, $swap;
			$fs2mp{$swap} = "swap";
		}
	}
}
##### sort_dffs sort table of mounts and swap from current os
#
# Input  @dffs (current fs in mount order + swap)
# Output @dffs sorted in partition # order.
# Note: this routine will need to be re-worked if 
# $MAXMOUNTS must support a value greater than 99.
#
sub sort_dffs {
	if ($#dffs > ($MAXMOUNTS-1)) {
		Pexit "ERR13: number of mount points > $MAXMOUNTS(maximum allowed)";
	}
	foreach $xxx (@dffs) {
		if ($xxx =~ m/([^\d]+)(\d)$/) {
		$xxx = $1 . "0" . $2;}
	}
	@sorted = sort @dffs;
	foreach $xxx (@dffs) {
		if ($xxx =~ m/([^\d]+)0(\d)$/) {
		$xxx = $1 . $2;}
	}
	foreach $xxx (@sorted) {
		if ($xxx =~ m/([^\d]+)0(\d)$/) {
		$xxx = $1 . $2;}
	}
}

#### build_partition_table_backup_os (Build Partition Table)
#
#	Build partition table for the backup drive -- by:
#  1. Copy verbatim the partition table for the current OS drive.
#                                              -- or:
#  2. Create a new partition table from user-supplied parameters.
#  The partition table created for the backup drive is formatted
#  to be input to the linux command, 'sfdisk', and includes the
#  creation of the 'extended partition' when there are to be more
#  than 4 total partitions on the backup drive.
#
#  Usage: 
#   If the 1st parameter is equal to the value "verbatim" then
#   the partitions on the backup drive will match in size type and
#   number the partitions on the current os drive.
#
#   If the 1st parameter is not equal to the string value "verbatim"
#   then it is assumed to be the total number of sectors per cylinder
#   available in the backup drive. the following are the args:
#	1. number of sectors/cylinder (on backup drive)
#	2. total number of cylinders  (on backup drive)
#   3-(N+2):  1 argument per new partition, each argument
#             containing size, partition type, offset, and bootable
####
sub build_partition_table_backup_os {
my $BFR;
my $format = "%-10s: start=%9u, size=%9u, Id=%.2s";
#1 GIGABYYE = 1073741824 bytes.
my $choice = $_[0];
my ($i,$j,$k);
my ($sz,$typ,$off,$boot);
my ($szp,$typp,$offp,$bootp);
if ($choice eq "verbatim") {
	runme "$GetHDpart $curr_os|grep -E '/dev|^unit'",
		  "ERR14:  called from build_partition_table_backup_os";
	@partbld = @cmdout;
	find_swap_curr_os;
	sort_dffs;
	foreach $devln (@sorted) {
		$mpt = $fs2mp{$devln};
		chomp $mpt;
		push @numpts, $mpt;
		$nufstyp{$mpt}="DEFAULT";
	}
	shift @partbld;
} else {
	@partbld = ();
	my $spc = shift @_;
	my $tcyl= shift @_;
	my $totsec = ($spc * $tcyl);
	my $sec    = $hddsec{$bkup_os};
	$i = 0; $j = 0; $k = scalar(@_);
	push @partbld, "unit: sectors";
	$offp = $sec;
	while ($pparm = shift @_) {
		($sz, $typ, $boot) = split /\//, $pparm;
		if ($sz)  {$szp = inp2sec $sz, $spc;}
		if ($szp < 1) {$szp = $totsec - $offp;}
		if ($typ) {$typp = $typ;}
		if ($boot) {$bootp = $boot;}
		if ($i > 0) {$szp = roundup($szp/$spc) * $spc;}
		else {$szp = roundup($szp/$spc) * $spc - $offp;}
		if ($j > 2 && $k > 1 && $i < 5) {
			my $tmp = $totsec - $offp;
			$i++; $j = 0;
			$BFR = sprintf  $format, "$bkup_os$i", $offp, $tmp, "f";
			push @partbld, $BFR;
			$offp += $sec; $szp -= $sec;
		} else {
			if ($i > 3) {$szp -= $sec;}
		}
		if ($szp lt 1) {$szp += ($totsec - $offp + $sec)}
		$i++; $j++; $k--;
		$BFR = sprintf  $format, "$bkup_os$i", $offp, $szp, $typp;
		if($bootp) {$BFR .= ", bootable";}
		push @partbld, $BFR;
		if ($i > 4) {$szp += $sec;}
		$offp += $szp;
		$bootp = "";
	}
}
return;
}

#### ck_nfs (Check for nfs mounts)
#
#  Verify that the designate backup drive does not have any
#  nfs mounted filesystems. Terminate with error message
#  there are nfs mounts.
####
sub ck_nfs {
my $nfscnt = 0;
runme "mount", "ERR15: called from ck_nfs";
foreach $mntln (@cmdout) {
if ($mntln =~ m/^\w+:/) {
	$nfscnt += 1;
	($mydev, $junk, $mymnt) = split / /, $mntln;
	print "ERR16: nfs/samba mount:  $mydev  mounted on:  $mymnt\n";
}
}
if ($nfscnt gt 0) {Pexit "ERR17: above nfs mount(s) need to be unmounted 
	prior to running 'mkbkup.pl'"}
}

#### ck_active_swap (Check for active swap partitions)
#
#  Verify that the designated backup drive does not have any
#  active swap partitions. Terminate with error code if thre
#  are one or more active swap partitions on the designated
#  backup drive.
####
sub ck_active_swap {
	runme "/sbin/swapon -s|grep $bkup_os|wc -l|awk '{print \$1}'";
	if ($cmdout[0] > 0) {
		Pexit "ERR18: Designated back up drive, '$bkup_os', has active swap partition(s)";
	}
}

#### ck_duppart (Check for Duplicate Partition)
#
#  Check for any partitions mounted on a common mountpoint.
#  If each partition does not have a separate unique mountpoint
#  then an error condition is assumed to exist and the mkbkup.pl
#  termiates.
####
sub ck_duppart {
my $nu_mtp= $_[0];
my $dev   = $_[1];
my $fstyp = $_[2];
	if (!exists($nump2fs{$nu_mtp})) {
		$nump2fs{$nu_mtp} = $dev;
		$nufs2fstyp{$dev} = $fstyp;
	} else {
		Pexit "ERR19: same mount point on 2 or more partitions
Mount Point =s $nu_mtp
Partition1  =s $nump2fs{$nu_mtp}
Partition2  =s $dev
Check runtime parms '--newpart $opt_newpart'
";
	}
}
#### mtpt_lvl (Mount-Point Level Check)
#
# Check for all filesystem mount points and make sure they are
# exactly one level deep.  Terminate with error if any fs 
# mount point is of the form: '/<dirname>/<subdir-name>'
#
####
sub mtpt_lvl {
my $sls_cnt = ($_[0] =~ s:/:/:g);
if ($sls_cnt > $MAXLEVELS) {
	Pexit "ERR20: in mtpt_lvl $_[0] is deeper than $MAXLEVELS levels";
}
return $sls_cnt;
}

sub prt_nufs2fstyp {
foreach $nu_mpt  (@numpts) {
	print "$nu_mpt with fstyp=$nufs2fstyp{$mp2fs{$nu_mpt}} on $mp2fs{$nu_mpt}\n";
}
}

#### tie_numpts_2_nuparts :: Tie New Mountpoints to New Partitions.
#
# Input: Parsed command line options specifying new partitions
#   mountpoints and sizes, in addition to data on current os.
# Output: filesystem types are determined for each new filesystem
#   which defaults to either $MAX_FSTYP or the fs type of the
#   corresponding old filesystem, if any.
# In addition:
# 1. this routine checks to see if 2 different partitions
#  mount to the same mount point which is treated as an error.
# 2. In the case in which new partitions are created, check to
#  see if the new partitions are large enough to back up existing
#  data.
#
sub tie_numpts_2_nuparts {
my ($nu_mtp, $dev, $curdu, $sz, $fstyp);
my $olddev;
my $i = 1;
my $totsz = 0;
foreach $nu_mtp (@numpts) {
	build_contains_fs_trail $nu_mtp;
	if ($partbld[$i] =~ m/(.+)(Id= f|Id=f)(.*)/) {$i++}
	$dev = $partbld[$i];
	chomp $dev;
	$sz = $dev;
	$sz =~ s/(.+)size=(\s*)(\d+)(.*)/$3/;
	$dev =~ s/(.+)\s*:(.+)/$1/;
	$dev =~ s:(\S+)\s*:$1:;
	if ($nufstyp{$nu_mtp} eq "DEFAULT") {
		if (exists($mp2fs{$nu_mtp})) {
			$olddev = $mp2fs{$nu_mtp};
			$fstyp = $fs2fstyp{$olddev};
		} else {$fstyp="$DEF_FSTYP";}
	} else { $fstyp=$nufstyp{$nu_mtp}; }
	if (!defined $opt_newpart) {
		if (exists($mp2fs{$nu_mtp})) {$dev = $mp2fs{$nu_mtp}}
	}
	$dev =~ s:$curr_os:$bkup_os:;
	ck_duppart $nu_mtp, $dev, $fstyp;
	$sz /= 2;
	if ($nu_mtp eq "swap") {
		$nufs2fstyp{$dev} = "N/A";
	}
	$sz = int $sz;
	if ($sz eq 0) {$sz = $hddsz{$bkup_os} - $totsz;}
	$nufscap{$nu_mtp} = $sz;
	$totsz += $sz;
	$i++;
}
if ($totsz >$hddsz{$bkup_os}) { 
	Pexit "ERR21:  $totsz KB (Total proposed partitions)
  -gt-$hddsz{$bkup_os} KB (capacity of backup drive '$bkup_os')\n";
}
print "$totsz(capacity required)  =< $hddsz{$bkup_os}(capacity of $bkup_os)\n";
}

#### build_new_partition_table_bkup_os :: Partition the Hard drive designated as the backup.
# Input: 
#   @partbld is the table of partitions to be built on the backup.
#   $bkup_os is the hard drive specified to be the backup.
# Output:
#   The hard drive specified by $bkup_os is re-partitioned.
#
sub build_new_partition_table_bkup_os {
my $repartition_cmd = "| $PutHDpart $bkup_os";
my $partln;
my $lncnt = scalar(@partbld) + 10;
print "===> Build a new partition table on '$bkup_os'.\n";
if (!(defined $opt_verbose)) {$repartition_cmd .= " >/dev/null 2>&1";}
open(REPART, $repartition_cmd)
or Pexit "ERR22: in 'build_new_partition_table_bkup_os', problem opening an output pipe";
foreach $partln (@partbld) {
	chomp $partln;
	print REPART "$partln\n";
}
close(REPART);
@partbld = ();
}
my $bb = "0123456789BCDFGHJKLMNPQRSTVWXYZbcdfghjklmnpqrstvwxyz/";
sub mungify {
my $num = $_[0];
my $nx = $_[1];
my ($mungln, $rm, $qn);
$ix = 0;
$qn = $num;
$mungln = "";
while ($ix<$nx) {
	$rm = $qn % 53;
	$qn /= 53;
	$qn =~ s/(.+)\.(.*)/$1/g;
	$mungln = substr($bb, $rm, 1) . $mungln;
	$ix++;
}
return $mungln;
}

# gen_a_lbl ### create a label for a fs which is used
#               by the 'LABEL=' parameter of mount commands.
# supporting subroutine for 'build_new_fs_labels'.
# Input
#   a prefix of 6 bytes or less.
#   process id and the current date and time.
# Output 
#   a label of 15 bytes or less that consists of the prefix,
#   plus a 9-byte suffix that contains process id and date
#   and time expressed as numbers to the base '53'.
#
sub gen_a_lbl {
#
# Create mungified microseconds (lo-order 3 bytes)
#
my $micro  = mungify `date +%N`/1000, 3;
#
# Create mungified process id (lo-order 2 bytes)
#
my $procid = mungify $$, 2;
#
# Create mungified yrs,julian,minutes date (4 bytes)
#
my $yr = `date +%Y`%13;
my $ju = `date +%j`;
my $sh = `date +%H`;
chomp $sh;
my $mn = `date +%M` + (60 * $sh);
chomp $mn;
chomp $yr;
chomp $ju;
my $yrju = ($yr * 366 + $ju)*1440 + $mn;
$yrju = mungify $yrju, 4;
return "$_[0]$yrju$micro$procid";
}

#### build_new_fs_labels :: Build volume/partition labels for backup hard drive
#   +++ NOTE: For 'ext3' filesystems ONLY !!!! +++
# Input: 
#   Current time and process id.
# Output:
#   partition/fs Labels on the hard drive at $bkup_os.
# Operation
#  All labels will be between 12 and 15 bytes in length
#  All labels are built using a base 53 number system
# using the digits 1-9, and the upper and lower case alphabet
# with all vowels removed to reduce  the possibility of
# cursing or other prurience, and the slash '/'.
#
sub build_new_fs_labels {
my ($inp, $pfx, $qqb);
print "===> Build new filesystem labels for '$bkup_os'.\n";
my $qq = 0;
my $sfx = gen_a_lbl("_");
my $nqqb = 1;
my $npfx = 5;
if ($#numpts > 53) {$nqqb=2;$npfx=4;}
foreach $mtp (@numpts) {
	if ($mtp eq "swap") {next;}
	$qqb = mungify($qq, $nqqb);
	if ($mtp ne "/") {
		$pfx = substr($mtp, 1, $npfx) . $qqb;
	} else {
		$pfx = "root" . $qqb;
	}
	$nulbl{$mtp} = $pfx . $sfx;
	$qq++;
}
}

sub FSprep {
my $fstyp = $_[0];
my $partn = $_[1];
my $mtpt  = $_[2];
my $lbl   = $_[3] unless(exists $nolbls{$fstyp});
print "mount=($partn" unless (!$partn);
print " $mtpt" unless (!$mtpt);
print " $fstyp" unless (!$fstyp);
print " $lbl" unless (!$lbl || (exists $nolbls{$fstyp}));
print ") ";
$res = `mount |grep $partn`;
($res) && Pexit "ERR23: $partn already mounted\nERR23: $res";
runme "/bin/mkdir -p $mtpt 2>&1";
runme "/bin/dd if=/dev/zero of=$partn count=100",
	"ERR24: error zeroing out 100 sectors of backup partition: $partn";
if ($fstyp eq "msdos" || $fstyp eq "vfat" ) {
	runme "/sbin/mkfs -t $fstyp $partn 2>&1",
		"ERR25: called from FSprep";
} else {
	if ($fstyp ne "reiserfs") {
		if ($lbl) {	
			runme "/sbin/mkfs -t $fstyp -L $lbl -q $partn 2>&1",
				  "ERR26: called from FSprep";
		} else {
			runme "/sbin/mkfs -t $fstyp -q $partn 2>&1",
				  "ERR27: called from FSprep";
		}
	} else {
		runme "/sbin/mkreiserfs -ff $partn 2>&1",
			  "ERR28: called from FSprep";
	}
}
runme "mount $partn $mtpt 2>&1", "ERR29: called from FSprep";
}

# update_grub_conf_bkup_os ... Create a grub configuration file for backup OS.
#
# Input: /boot/grub/grub.conf
# Output: <backup-drive-root-mount-point>/boot/grub/grub.conf
#
# Description: Using the grub.conf file on the current system as
#		a template, create a grub.conf file for the backup hard
#		drive so that the backup harddrive will boot properly.
#
sub update_grub_conf_bkup_os {
runme "df $bkup_mtpt/boot|tail -1",
	  "ERR30: called from update_grub_conf_bkup_os";
$bootpn = $cmdout[0];
chomp $bootpn;
$bootpn =~ s/([\/.]*)[ \t]{1,99}(.*)/$1/;
$bootpn =~ s/\D+(\d+)$/$1/;
$bootpn -= 1;
runme "df |grep '$bkup_mtpt/boot' |wc -l|awk '{print \$1}'",
	  "ERR31: called from update_grub_conf_bkup_os";
chomp $cmdout[0];
if ($cmdout[0] eq "0") {$bootdir="/boot/";}
else {$bootdir="/";}
$bkup_grub_conf = "$bkup_mtpt/boot/grub/grub.conf";
runme "cat /boot/grub/grub.conf",
	  "ERR32: called from update_grub_conf_bkup_os";
foreach $grubln (@cmdout) {
	if ($grubln =~ m/^#/) {next;}
	if ($grubln =~ m/root=/) {
		$grubln =~ s/(.+root=)(.*)/$1/;
		chomp $grubln;
		$dev = $nump2fs{"/"};
		$fstyp = $nufs2fstyp{$dev};
		$dev =~ s:$bkup_os:$curr_os:;
		if (exists $nolbls{$fstyp}) { $grubln .= $dev; }
		else                        { $grubln .= "LABEL=$nulbl{'/'}"; }
		$grubln .= "\n";
	}
	$grubln =~ s/\(hd0,.\)/\(hd0,$bootpn\)/;
	$grubln =~ s/ (\/boot)*\// $bootdir/;
	$grubln =~ s/\)(\/boot)*\//\)$bootdir/;
}
my @loopgrub = @cmdout;
open (NUGRUB, ">$bkup_grub_conf") or 
	Pexit "ERR33: Can't open $bkup_grub_conf for writing";
foreach $partln (@loopgrub) {print NUGRUB "$partln";}
close(NUGRUB);
if (defined $opt_verbose) {
	print "### 'grub.conf' as editted by mkbkup.pl\n";
	open (NUGRUB, "<$bkup_grub_conf") or 
		Pexit "ERR34: Can't open $bkup_grub_conf for reading";
	foreach $partln (<NUGRUB>) {print "$partln";}
	close(NUGRUB);
}
}

# make_grub_device_map_bkup_os ... Create a device.map file 
#
# Input: $bkup_os, $curr_os
# Output: $bkup_mtpt/boot/grub/device.map
#
# Description: create a device.map file for the backup hard drive
#    that is aware of both the current OS drive and the backup drive.
#
sub make_grub_device_map_bkup_os {
my $format = "%-8s %s";
open (NUGRUB, ">$bkup_mtpt/boot/grub/device.map") or 
	Pexit "ERR35: Problem opening $bkup_mtpt/boot/grub/device.map for writing";
	print NUGRUB "#device.map generated by: 'mkbkup.pl'\n";
	print NUGRUB "(fd0) /dev/fd0\n";
	print NUGRUB "(hd0) $curr_os\n";
	print NUGRUB "(hd1) $bkup_os\n";
	close(NUGRUB);
}

# make_mbr_bkup_os ...  Create master boot record on backup device
#
# Input: $bkup_os, $bkup_mtpt
# Output: $bkup_os... track 0 sector 0 cylinder 0
# Calls:  make_grub_device_map_bkup_os
#         update_grub_conf_bkup_os
# Description: Create the master boot record on the backup device.
#
sub make_mbr_bkup_os {
	make_grub_device_map_bkup_os;
	update_grub_conf_bkup_os;
	runme "/sbin/grub-install --root-directory=$bkup_mtpt $bkup_os",
		  "ERR36: called from make_mbr_bkup_os";
	if (defined $opt_verbose) {prt_cmdout;}
}

# make_new_dir ... give a mount point specific permissions
#                 after a fs is mounted on that mount point.
#                 involves creating a temporary directory in
#                 safe place (one not being backed up), changing
#                 the permissions on that directory, then using
#                 'cp -pdR <dir> <parent-of-mount-point>' to 
#                 change permissions on the mount point post-mount.
#  Input: 1 argument is relative pathname of mount point that
#         has specific permissions to be changed.
#  Output: fs has permissions of root node changed.
#
sub make_new_dir {
my $newpath = "$bkup_mtpt$_[0]";
my $tmppath = "$bkup_mtpt$bkup_mtpt$_[0]";
my $parentdir = dirname $_[0];
runme "basename $_[0]";
my $basedir = $cmdout[0];
chomp $basedir;
runme "ls -l $parentdir|grep $basedir|awk '{print \$1}'";
my $oldperms = $cmdout[0];
chomp $oldperms;
runme "mkdir -p $tmppath";
my $cmd = "chmod " . chmode($oldperms) . " $tmppath";
runme "$cmd";
runme "cp -pdR $tmppath $bkup_mtpt$parentdir";
runme "rm -fr $bkup_mtpt$bkup_mtpt";
}

my $rlvl=0;

# enter_mountpoint_subdir 
#     Enter a subdirectory containing a fs mount point and
#     continue the copy to backup device.  This allows 
#     filesystem mount points to be more than 1 level deep
#     from / (root) filesystem.  e.g., '/usr/local' could
#     be a mount point for a filesystem.
#
# Input:
#     1st argument is the fullpath of the subdirectory.
# Output:
#     continuing of copy to backup device.  
#     mount filesystems at subdirectory level when applicable.
#
sub enter_mountpoint_subdir {
my $mypath = $_[0];
my $curr_fs;
my @looproot = (getdir $mypath);
make_new_dir $mypath;
Pexit "ERR37: Can't go more than $MAXLEVELS deep!" unless ($rlvl < $MAXLEVELS);
$rlvl += 1;
foreach $blurb (@looproot) {
	$path = $mypath . "/" . $blurb;
	$cpcmd = "cp -pdR $path $bkup_mtpt$mypath";
	if (exists($nump2fs{$path})) {
		$curr_fs = $nump2fs{$path};
		$fstyp = $nufs2fstyp{$nump2fs{$path}};
 		if ($nufstyp{$path} ne "DEFAULT") {$fstyp = $nufstyp{$path}};
		FSprep $fstyp, $nump2fs{$path}, "$bkup_mtpt$path", $nulbl{$path};
		print "\n";
	}
	print substr("            ",0,($rlvl*2));
	print "copy $path\n";
	if (exists($contains_fs{$path})) {
		enter_mountpoint_subdir($path);
	} else {
		runme "$cpcmd";}
}
$rlvl -= 1;
}

# make_new_fs_on_bkup_os == make new filesystems on backup disk drive
#
# Create new filesystems for the backup os and mount them in 
# hierarchical order on a designated root mountpoint and copy
# from the current OS to these new filesystems.
# 
# Input: 1. All MOUNTED filesystems on the current OS.
#        2. Command line options.
#        3. the partition table of the backup OS.
# Output: A bootable backup of the current OS on a different
#    designated hard drive.
#
sub make_new_fs_on_bkup_os {
my ($cpcmd, $fstyp, $curr_fs, $path);
print "===> Build new filesystems on '$bkup_os'.\n";
$fstyp = $nufs2fstyp{$nump2fs{"/"}};
if ($nufstyp{"/"} ne "DEFAULT") {$fstyp = $nufstyp{"/"}};
$curr_fs = $nump2fs{"/"};
FSprep $fstyp, $nump2fs{"/"}, "$bkup_mtpt", $nulbl{"/"};
chdir "/" or Pexit "ERR38: Could not 'chdir /'";
my @looproot = (getdir "/");
foreach $blurb (@looproot) {
	$cpcmd = "cp -pdR $blurb $bkup_mtpt";
	$path = "/" . $blurb;
	next unless($path ne $bkup_mtpt);
	if ("/proc" eq $path) {
		runme "mkdir $bkup_mtpt/proc",
			  "ERR39: called from make_new_fs_on_bkup_os";
		runme "chmod 555 $bkup_mtpt/proc",
			  "ERR40: called from make_new_fs_on_bkup_os";
		next;
	}
	if (exists($nump2fs{$path})) {
		$curr_fs = $nump2fs{$path};
		$fstyp = $nufs2fstyp{$nump2fs{$path}};
 		if ($nufstyp{$path} ne "DEFAULT") {$fstyp = $nufstyp{$path}};
		FSprep $fstyp, $nump2fs{$path}, "$bkup_mtpt$path", $nulbl{$path};
	}
	print "copy $path to $curr_fs\n";
	$curr_fs = $nump2fs{"/"};
#  	if (($path ne "/etc") && ($path ne "/boot") &&
# 			($path ne "/lvl1") && ($path ne "/spare2G")) #for DEBUG SPEED UP uncomment
# 		{next} #for DEBUG SPEED UP uncomment
	if (exists ($contains_fs{$path})) {enter_mountpoint_subdir($path);}
	else                              {runme "$cpcmd";}
}
}

sub make_swap_fs_bkup_os {
Pexit "ERR41: SWAP NOT DEFINED" unless($_[0]);
my $swap = $_[0];
print "===> Do a '/sbin/mkswap $swap'.\n";
print "swap partition =s $swap!\n";
runme "/sbin/mkswap $swap",
	  "ERR42: called from make_swap_fs_bkup_os";
prt_cmdout;
}

# update_etc_fstab_bkup_os Build the etc/fstab for the backup disk
#
# Based on what is already mounted at the backup root point
# and backup root mount point subdirectories, create an etc/fstab
# for the backup disk.
#
# OUTPUT: The new etc/fstab is built as follows:
#  1. a line in new etc/fstab for every mounted backup filesystem.
#  2. no swap entries for current os are copied to new one.
#  3. One swap partition is added for the backup swap partition created.
#  4. remaining non-filesystem mount points are copied from existing 
#     /etc/fstab.
# NOTE: the above order is preserved in the new backup etc/fstab.
# INPUT:
#  1. /etc/fstab from current os.
#  2. system info about all mounted backup filesystems.
#  3. information about the newly created backup swap partition.
#
sub update_etc_fstab_bkup_os { ### build the 'fstab' based on what is already mounted.
my @nu_fstab;
my ($dev, $mpt, $mtyp, $mopts, $mx1, $mx2);
my $format = "%-23s %-16s %-8s %-17s %s %s";
chdir "/etc"		or 
	Pexit "ERR43: Couldn't chdir to /etc";
#
# create the mounts for the new filesystems, incl. / and /boot
#
runme "mount |grep $bkup_os",
	  "ERR44: called from update_etc_fstab_bkup_os";
foreach $fstbln (@cmdout) {
	($dev, $junk, $mpt, $junk, $mtyp, $junk) = split / /, $fstbln;
	$mopts = "defaults";
	$mpt =~ s:$bkup_mtpt(.*):$1:;
	$dev =~ s:$bkup_os:$curr_os:;
	$mx1=1; $mx2=2;
	if (!$mpt) {$mx2=1; $mpt="/";}
	if (exists($nulbl{$mpt})) {
		if ($mtyp =~ m/ext/) {$dev = "LABEL=$nulbl{$mpt}";}
	}
	$fstbln = sprintf  $format, $dev, $mpt, $mtyp, $mopts, $mx1, $mx2;
	push @nu_fstab, "$fstbln\n";
}
#
# output the swap partition here ...
#
if (exists $nump2fs{"swap"}) {
	$fstbln = sprintf $format, $nump2fs{"swap"}, "swap", "swap", "defaults", 0, 0;
	$fstbln =~ s:$bkup_os:$curr_os:;
	push @nu_fstab, "$fstbln\n";
}
#
# copy remaining mounts from /etc/fstab
#
runme "cat /etc/fstab|grep -v \"1 \[12\]\$\"",
			  "ERR45: called from make_new_fs_on_bkup_os";
foreach $fstbln (@cmdout)  {
	$fstbln =~ tr/\t/ /;
	$fstbln =~ s/ {1,99}/ /g;
	($dev, $mpt, $mtyp, $mopts, $mx1, $mx2) = split / /, $fstbln;
	next unless ($mtyp ne "swap");
	$fstbln = sprintf  $format, $dev, $mpt, $mtyp, $mopts, $mx1, $mx2;
	push @nu_fstab, "$fstbln";
}
if (defined $opt_verbose) {
	print "### $bkup_mtpt/etc/fstab created for the bootable backup on $bkup_os\n";
	foreach $fstbln (@nu_fstab) {
		print $fstbln;
	}
}
open (NUFSTAB, ">$bkup_mtpt/etc/fstab") or 
	Pexit "ERR46: Problem opening $bkup_mtpt/etc/fstab for writing";
foreach $fstbln (@nu_fstab) {
	print NUFSTAB "$fstbln";
}
close(NUFSTAB);
}

# unmount_bkup_fs ... unmount all filesystems on the bootable backup.
#   A unmount loop is started, which attempts to unmount the
#   filesystem with the longest path name.
# Input: All mounted filesystems on the backup drive.
# Output: All filesystems on backup drive are unmounted.
#
sub unmount_bkup_fs {
my ($dev,$mtpt,$junk,$mxmtpt,$mxlen,$mxdev); 
my $mtcnt = 0;
$mxlen = 0;
runme "cd /;sync;sync";
while (1 eq 1) {
	runme "mount", "ERR47: called from unmount_bkup_fs";
	$mxlen = 0; $mtcnt = 0;
	foreach $mtln (@cmdout) {
		next unless ($mtln =~ m:$bkup_os:);	
		$mtcnt += 1;
		($dev, $junk, $mtpt, $junk, $junk) = split / /, $mtln;
		if ($mxlen<length($mtpt)) {
			$mxlen = length($mtpt);
			$mxmtpt = $mtpt;
			$mxdev = $dev;
		}
	}
	return unless ($mtcnt>0);
	runme "sleep 2;umount $mxmtpt", "ERR48: unmount_bkup_fs";
	print "'$mxmtpt' on '$mxdev' successfully umounted.\n";
}
}

sub NoBackupDrive {
	print "No Backup Harddrive was Defined.\n";
	runme "$GetHDgeom | awk '{print \$1}'", "ERR49: called from NoBackupDrive";
	prt_cmdout;
	print "Pick from above list, but NOT the current drive!!\n";
	exit (0);
}

# fs_check_capacity === make sure new partitions are of adequate
#                    capacity to hold corresponding directory trees.
# Input: actual amount of bytes in use for each directory tree or subtree
#        that will be a filesystem in a partition on the backup drive.
# Output: display bytes in use vs. fs capacity for each fs that will exist
#        on the backup drive.
# Outcome:
#     continue if every partition on backup drive has sufficient capacity.
#     terminate program with an error message if any partition is too small.
#
sub fs_check_capacity {
my $szA=0;
my ($szs, $mpt, $pfx, $junk, $fsCAP);
my $format = "%-15s %9sKB %9sKB";
my %nufs_size = ();
#
# Grab existing sizes .........
#
print "===> Check space requirements of new filesystems ... Please wait\n";
if (defined $opt_verbose) {
print "    (Checking ALL filesystems)...";
}
if (defined $opt_debug) {print "\n";}
foreach $dfent (@dffs) {
	$mpt = $fs2mp{$dfent};
	if ($mpt eq "swap") {next}
	if ($mpt eq "/") {
		@cmdout = `df /|tail -1|awk '{print \$3}'`;
	} else {
		if (defined $opt_verbose && !(defined $opt_debug)) {print ",";}
			runme "du --summarize $mpt",
		      "ERR50: FATAL ...should have been able to find '$mpt'!\n";
	}
	$szs = $cmdout[0];
	$szs =~ s:(\d+)[^\d].+:$1:;
	chomp $szs;
	$nufs_size{$mpt} = $szs;
	$szA += $szs;
	if (defined $opt_verbose) {print "$mpt";}
	if (defined $opt_debug) {print "=$nufs_size{$mpt}\n";}
}
$nufs_size{"/"} = $szA;
#
# Collect sizes of proposed new partitions (may be old)
#
foreach $mpt (@numpts) {
	next unless($mpt ne "/");
	next unless($mpt ne "swap");
	if (!exists($nufs_size{$mpt})) {
		if (defined $opt_verbose && !(defined $opt_debug)) {print ",";}
			runme "du --summarize $mpt",
		      "ERR51: '--newpart' parameter-string error ... could not find '$mpt'.";
	    $cmdout[0] =~ s:(\d+)[^\d].+:$1:;
		$nufs_size{$mpt} =  $cmdout[0];
		chomp $nufs_size{$mpt};
		if (defined $opt_verbose) {print "$mpt";}
		if (defined $opt_debug) {print "=$nufs_size{$mpt}\n";}
	}
	$pfx = 100 - mtpt_lvl($mpt); ### create prefixed entry with level number
	if($pfx<10) {$pfx = "0$pfx";}
	push @pfxdmtp, "$pfx$mpt";
}
print "\n";
#
# Sort the array of prefixed mount points in level-number order
#
@pfxdmtp = sort @pfxdmtp;
push @pfxdmtp, "00/";
#
# go thru the sorted list to calculate actual fs kilobytes in use.
#
if (defined $opt_verbose) {
	print "new filesystem       in-use    capacity\n";
}
foreach $prt (@pfxdmtp) {
	$mpt = substr($prt,2);
	$BFR = sprintf  $format, $mpt, $nufs_size{$mpt}, $nufscap{$mpt};
	if (defined $opt_verbose) {print $BFR, "\n";}
	if ($nufs_size{$mpt} > ($nufscap{$mpt} - 1000)) {
		print "...CAPACITY EXCEEDED\n";
		Pexit "ERR52: Capacity of fs='$nump2fs{$mpt}' INSUFFICIENT to hold '$mpt'";
	}
	$szs = $nufs_size{$mpt};
	while ($mpt ne "/") {	# adjust all parent filesystem sizes downward.
		$mpt = dirname $mpt;
		if (exists $nufs_size{$mpt}) {
			$nufs_size{$mpt} -= $szs;
		}
	}
}
}

###########################################################################
if ($ENV{DEBUG}) {
	print "DEBUG:mkbkup.pl VERSION: $VERSION\n";
}
use Getopt::Long;;
use Term::ANSIColor qw(:constants);


############# Command Line options processing and default values ##################
#
# If you want to change the default behavior of this script, modify the values here
# Please notice that for some of them, environment variables are honored.

my $rc_getoptions = GetOptions("debug" => \$opt_debug,
           "newpart=s" => \$opt_newpart,
           "output=s" => \$opt_output,
           "nolabels=s" => \$opt_nolabels,
           "verbose" => \$opt_verbose,
           "help" => \$opt_help,
);###GetOptions
if (!$rc_getoptions) {
	Pexit "ERR53: Invalid option passed to `basename $0`";
}

if (defined $opt_help) {
	&usage;
}
if (defined $opt_output) {
	$bkup_os = $opt_output;
}

if (!defined $bkup_os) {$bkup_os = shift or NoBackupDrive;}
if ($bkup_os =~ m:^dev/:) {
	$bkup_os = "/" . $bkup_os;
}
if (!($bkup_os =~ m:/:)) {
	$bkup_os = "/dev/" . $bkup_os;
}
if (defined $opt_nolabels) {
	print "fstypes which don't support labels in kernel = $opt_nolabels\n";
	foreach $badboy (split /,/, $opt_nolabels) {
		$nolbls{$badboy} = "NOLABEL";
	}
	if ((exists $nolbls{"ALL"}) || (exists $nolbls{"all"})) {
		$nolbls{"ext2"}="NOLABEL";
		$nolbls{"ext3"}="NOLABEL";
		$nolbls{"vfat"}="NOLABEL";
		$nolbls{"msdos"}="NOLABEL";
		$nolbls{"jfs"}="NOLABEL";
		$nolbls{"xfs"}="NOLABEL";
		$nolbls{"reiserfs"}="NOLABEL";
	}
}

$bkup_mtpt="/bkupmp_$$";
runme "mkdir $bkup_mtpt", "ERR54: backup mountpoint '$bkup_mtpt' already exists",
                          "       either delete it or rename it or run '$0' again";

################### BEGIN TEST SECTION
#$bkup_os = "/dev/hde";
#$curr_os = "/dev/hde";
#$nulbl{'/'} = "root_nnnn";
#update_grub_conf_bkup_os;
#my $cmd;
#foreach $x1x ("--", "-w", "r-", "rw") {
#foreach $x1y ("-", "x", "s", "S") {
#foreach $x2x ("--", "-w", "r-", "rw") {
#foreach $x2y ("-", "x", "s", "S") {
#foreach $x3x ("--", "-w", "r-", "rw") {
#foreach $x3y ("-", "x", "t", "T") {
#$chmod_inp = "d" . $x1x . $x1y . $x2x . $x2y . $x3x . $x3y;
#$cmd = "chmod " . chmode($chmod_inp) . " tmp_tmp";
#runme "$cmd;ls -l|grep tmp_tmp|awk '{print \$1}'";
#chomp $cmdout[0];
#if ($cmdout[0] ne $chmod_inp) { Pexit "$chmod_inp => $cmd  ==> $cmdout[0]!!!!!!\n";}
#else {print "."}
#}}}}}}
#build_contains_fs_trail "///a/b/c";
#build_contains_fs_trail "/level1";
#print "\$contains_fs{\"/a/b\"} = $contains_fs{'/a/b'}\n";
#print "\$contains_fs{\"/a\"} = $contains_fs{'/a'}\n";
#print "\$contains_fs{\"/a/b/c\"} = $contains_fs{'/a/b/c'}\n";
#print "\$contains_fs{\"/\"} = $contains_fs{'/'}\n";
################### END TEST SECTION

if (defined $opt_debug) {$opt_verbose=1;} # debug option includes verbose option

print_params;

print "Uid = $<, Eff Uid = $>\n";

Pexit "ERR55: Must run 'mkbkup.pl' as root" unless (!$>);
########### END OF Command Line Options ###############

#
# === Main Program consists mainly of subroutine calls
#

build_disk_geometry_table;

build_disk_size_table;

if (exists($hddsz{$bkup_os})) {
	print "Proposed Backup HDD : $bkup_os\n";
} else {
	Pexit "ERR56: '$bkup_os', proposed backup hdd, not found";
}

ck_nfs;			### check for any nfs mounts

ck_active_swap;	### check for active swap partitions on backup drive

build_dev_mount_table_curr_os;

if (defined $opt_newpart) {
	parse_nupart;
}

if ($bkup_os eq $curr_os) {
	Pexit "ERR57: proposed backup HDD($bkup_os) =s Current OS ($curr_os)
(you probably don't want to do this so I won't LET YOU!)";
}

if (! defined $opt_newpart) {
	print "Using the original partitions\n";
	build_partition_table_backup_os "verbatim";
} else {
	print "Using new partitions\n";
	build_partition_table_backup_os $secpcyl{$bkup_os}, $hddcyl{$bkup_os}, @partinp;
}

print "===> Check that new partitions will not exceed capacity of $bkup_os\n";
tie_numpts_2_nuparts;

my $format = "%-14s %-10s %-7s\n";
if (defined $opt_verbose) {
	print "================= CURRENT OS HDD : $curr_os\n";
	print sprintf $format, "mount-point", "partition", "fs-type";
	foreach $devln (@mpts) {
		print sprintf $format, $devln, $mp2fs{$devln}, $fs2fstyp{$mp2fs{$devln}};
	}
}
print "================= BACKUP  OS HDD : $bkup_os\n";
print sprintf $format, "mount-point", "partition", "fs-type";
foreach $devln (@numpts) {
	print sprintf $format, $devln, $nump2fs{$devln}, $nufs2fstyp{$nump2fs{$devln}};
}

prt_hdd_tbl;

build_new_fs_labels;

if (defined $opt_debug) {prt_nulbl}
 
if (defined $opt_newpart) {fs_check_capacity}

build_new_partition_table_bkup_os;

if (exists $nump2fs{"swap"}) {
	make_swap_fs_bkup_os $nump2fs{"swap"};
}

make_new_fs_on_bkup_os;

make_mbr_bkup_os;

update_etc_fstab_bkup_os; ### new labels for all /etc/fstab mounted fs's.

unmount_bkup_fs;
runme "rmdir $bkup_mtpt", "ERR58: error while removing temporary mount point";
